local addonName, addonData = ...

local addonVersion = GetAddOnMetadata("Monochrome", "Version") 

local protocolDelimiter = "\1"

local lootlist = {} -- String encoded NAME:GUID
local lootlistMap = {}
local playersWithAddon = {}
local itemTrackingTable = {}
local similarItemIds = {}
local addonMessagePrefix = "MONO"
local lastBossKilled = nil

local events = {}
local addonMessageHandlers = {}

-----------------
-- Protocol
-----------------

local function buildMessage(protocol, param1, param2, param3)
    local msg = protocol
    if param1 then
        msg = msg .. protocolDelimiter .. param1
    else
        msg = msg .. protocolDelimiter
    end
    if param2 then
        msg = msg .. protocolDelimiter .. param2
    else
        msg = msg .. protocolDelimiter
    end
    if param3 then
        msg = msg .. protocolDelimiter .. param3
    else
        msg = msg .. protocolDelimiter
    end
    
	return msg
end

local function sendAddonMessage(protocol, param1, param2, param3)    
    C_ChatInfo.SendAddonMessage(addonMessagePrefix, buildMessage(protocol, param1, param2, param3), "RAID")
end

local function sendAddonWhisper(player, protocol, param1, param2)
    C_ChatInfo.SendAddonMessage(addonMessagePrefix, buildMessage(protocol, param1, param2), "WHISPER", player)
end

local function parseMessage(message)
	local split = addonData.stringSplit(message, protocolDelimiter)
	return split[1], split[2], split[3], split[4]
end

-----------------------------
-- addonMessageHandlers table
-----------------------------

local function isRaidLeadOrAssist(sender)
    local raidIndex = UnitInRaid(sender)
    local _, rank = GetRaidRosterInfo(raidIndex) -- 1 = promoted, 2 = raid leader
    return rank > 0
end

function addonMessageHandlers:DISCOVER(sender)
    sendAddonWhisper(sender, "REPORT_VER", addonVersion)
end

function addonMessageHandlers:REPORT_VER(sender, addonVersion)
    table.insert(playersWithAddon, {player = sender, version = addonVersion})
end

function addonMessageHandlers:LOOT_DETECTED(sender, guid, itemLink)
    local itemSplit = addonData.stringSplit(itemLink, ':')
    local itemId = itemSplit[2]
    local hasItemValue = false
            
    if lootlistMap[guid] then
        for index, value in ipairs(lootlistMap[guid].loot) do
            local splitResult = addonData.stringSplit(value, ':')
            if splitResult[2] == itemId then
                hasItemValue = true
            end
        end

        if not hasItemValue then
            table.insert(lootlistMap[guid].loot, itemLink)
        end
    else
        lootlistMap[guid] = { player = "?", loot = { itemLink } }
    end    
end

function addonMessageHandlers:DISCOVER_ITEM(sender, location, itemId)
    itemId = tonumber(itemId)
    local itemCount = 0
    if location == "equipped" then
        if IsEquippedItem(itemId) then
            itemCount = 1
        end
    else
        itemCount = GetItemCount(itemId, false)
        if similarItemIds[itemId] then
            for i,v in ipairs(similarItemIds[itemId]) do
                itemCount = itemCount + GetItemCount(v, false)
            end
        end
    end

    local item = Item:CreateFromItemID(itemId);
    item:ContinueOnItemLoad(function ()
        local itemName, itemLink = GetItemInfo(itemId)

        if not isRaidLeadOrAssist(sender) then
            sendAddonWhisper(sender, "REPORT_INSPECTION", "BLOCKED")
        else
            --addonData.printColor(sender .. " hat dich inspiziert für " .. itemLink)
            sendAddonWhisper(sender, "REPORT_INSPECTION", itemCount)
        end
    end)
end

function addonMessageHandlers:REPORT_INSPECTION(sender, itemCount)   
    table.insert(itemTrackingTable, { player = sender, quantity = itemCount })
end

function addonMessageHandlers:CAN_LOOT(sender, creatureName, creatureGuid)
    local text = creatureName .. ":" .. creatureGuid
    if lootlist[text] then
        lootlist[text] = lootlist[text] + 1
    else
        lootlist[text] = 1
        
        addonData.callback(1, function()
            local _, mlPlayerId = GetLootMethod();
            local playerIsMasterLooter = mlPlayerId == 0
			local npcId = addonData.getNpcId(creatureGuid)

            if lootlist[text] > 1 then
                lootlistMap[creatureGuid] = { player = "mehrere Spieler", loot = {} }
            elseif lootlist[text] == 1 then
                lootlistMap[creatureGuid] = { player = sender, loot = {} }

                local skinningTargetIds = {
                    11671, -- Core hound
                    11673, -- Ancient Core Hound
                }
                
                if addonData.hasValue(skinningTargetIds, npcId) then
                    addonData.printColor(sender .. " kann " .. creatureName .. " looten.")

                    if playerIsMasterLooter then
                        addonData.sendWhisperMessage(sender, "[Monochrome]: Bitte loote " .. creatureName .. " (gerade gestorben)")
                    end
                end
            end
        end)
    end
end

function addonMessageHandlers:SET_MT(sender, name1, name2, name3)
    if not addonData.savedVariablesAvailable then
        print("Addon noch nicht initialisert, bitte warte einen Moment.")
        return
    end

    if not addonData.removeValue(addonData.tanks, name1) then
        print(name1 .. " nicht gefunden")
    end
    
    if name2 then
        if not addonData.removeValue(addonData.tanks, name2) then
            print(name2 .. " nicht gefunden")
        end
    end
    
    if name3 then
        if not addonData.removeValue(addonData.tanks, name3) then
            print(name3 .. " nicht gefunden")
        end
    end
    
    if name3 then
        table.insert(addonData.tanks, 1, name3)
    end
    
    if name2 then
        table.insert(addonData.tanks, 1, name2)
    end
    
    table.insert(addonData.tanks, 1, name1)

    print((name1 or "invalid") .. "/" .. (name2 or "invalid") .. "/" .. (name3 or "invalid") .. " als MT gesetzt")
    
    MonoOptions.mt1 = name1
    MonoOptions.mt2 = name2
    MonoOptions.mt3 = name3
end

----------------
-- events table
----------------

function events:ADDON_LOADED(arg1)
    if arg1 ~= "Monochrome" then
        return
    end
    
    if MonoOptions == nil then
        MonoOptions = {}
    end
    
    addonData.savedVariablesAvailable = true
    
    if MonoOptions.mt1 ~= nil then
        addonData.setMaintank(MonoOptions.mt1, MonoOptions.mt2, MonoOptions.mt3)
    end 
end

function events:CHAT_MSG_ADDON(prefix, text, channel, sender, target)
    if prefix ~= addonMessagePrefix or channel == "SAY" or channel == "YELL" or channel == "GUILD" or not IsInRaid() then
        return
    end
    
    local protocol, param1, param2, param3 = parseMessage(text)
	param1 = param1 or "nil"
	param2 = param2 or "nil"
    param3 = param3 or "nil"

    local handler = addonMessageHandlers[protocol]
    if not handler then
        return
    end
    
    sender = addonData.playerNameRemoveRealm(sender)
    handler(self, sender, param1, param2, param3)
end

function events:PLAYER_LOGIN()
    C_ChatInfo.RegisterAddonMessagePrefix(addonMessagePrefix)
end

function events:LOOT_OPENED()
	if not IsInRaid() then
		return
	end
    local guid = UnitGUID("target")

    if not guid or not UnitIsDead("target") then
        return
    end
    
    if not lootlistMap[guid] then
        lootlistMap[guid] = { player = "?", loot = {} }
    end

    for i = 1, GetNumLootItems() do
        if LootSlotHasItem(i) then
            local itemLink = GetLootSlotLink(i)
            if itemLink then
                local _, _, _, _, rarity = GetLootSlotInfo(i);

                if rarity >= GetLootThreshold() then
                    sendAddonMessage("LOOT_DETECTED", guid, itemLink)
                end
            end
        end
    end
end

function events:UPDATE_MOUSEOVER_UNIT()
	if not IsInRaid() then
		return
	end
    local unitGuid = UnitGUID("mouseover")
    
    if not UnitIsDead("mouseover") or not lootlistMap[unitGuid] then
        return
    end
    
    GameTooltip:AddLine("Lootbar von " .. lootlistMap[unitGuid].player)

    if next(lootlistMap[unitGuid].loot) then
        local hasLoot, _ = CanLootUnit(unitGuid);

        if hasLoot then
            table.foreach(lootlistMap[unitGuid].loot, function (k, v)
                GameTooltip:AddLine(v)
            end)
        else
            lootlistMap[unitGuid].loot = {}
        end
    end
    GameTooltip:Show()
end

local blackList =
{
	17877, 18871, 18870, 18868, 18867, 18869, --shadowburn
	172, 25311, 11672,	6223, 11671, 6222, 7648, --corruption
	14323, 14324, 14325, 1130, --huntersmark
	1978, 13549, 13551, 13552, 13553, 13554, 13555, 25295, --serpentsting
	5570, 24974, 24975, 24976, 24977, --insect swarm
	16511, 17347, 17348, --hemorrhage
	8921, 8924, 9835, 9834, 8926, 8925, 8928, 8927, 9833, 8929, --moonfire
	12294, 21551, 21552, 21553, --mortal strike
	15286, --vampiric embrace
	3034, 14279, 14280, --viper sting
	3043, 14277, 14276, 14275, --scorpid sting
	2835, 2837, 11357, 11358, 25347, --deadly poison... yeah ok, only works when boss is targeted while poison is applied
	702, 1108, 6205, 7646, 11707, 11708, --curse of weakness
    1120, 8288, 8289, 11675, --drain soul
}

local currentCastGuid = nil

function events:UNIT_SPELLCAST_SENT(unit, target, castGUID, spellID)
    if not unit == "player" then
        return
    end
    currentCastGuid = castGUID
end

function events:UNIT_SPELLCAST_SUCCEEDED(unitTarget, castGuid, spellId)
	if not IsInRaid() then
		return
	end
    
	-- local targetGuid = UnitGUID("target")

	-- if targetGuid and castGuid == currentCastGuid and addonData.hasValue(blackList, spellId) then
        -- local currentCastGuid = nil -- this should help prevent chat spam when UNIT_SPELLCAST_SUCCEEDED is triggered twice
		-- local healthPercent = UnitHealth("target") / UnitHealthMax("target")
		-- local npcId = addonData.getNpcId(targetGuid)
		-- local onyxia = 10184
        -- local veklor = 15276
		-- if healthPercent > 0.06 and npcId ~= onyxia and npcId ~= veklor and addonData.isBoss(npcId) then
			-- if not addonData.isAllowed(UnitName("player"), spellId) then
				-- local spellName = GetSpellInfo(spellId)
				-- addonData.sendChatMessage("[Monochrome]: Sorry, ich habe " .. spellName .. " benutzt, kommt nicht wieder vor!")
			-- end
		-- end
	-- end
	
    local dowseRuneSpellId = 21358
    if unitTarget == "player" and spellId == dowseRuneSpellId then
        local chatMsg = "[Monochrome]: Ich habe eine Rune gelöscht"
        if lastBossKilled then
            chatMsg = chatMsg .. " (" .. lastBossKilled .. ")"
        end
        addonData.sendChatMessage(chatMsg)
    end
end

function events:COMBAT_LOG_EVENT_UNFILTERED()
	if not IsInRaid() then
		return
	end
    local _, eventName, _, _, _, _, _, destGuid, destName = CombatLogGetCurrentEventInfo()
    local isCreature = addonData.startsWith(destGuid, "Creature")
    if eventName ~= "UNIT_DIED" or not isCreature then
        return
    end
    
    local npcId = addonData.getNpcId(destGuid)
    if npcId and addonData.isBoss(npcId) then
        lastBossKilled = destName
    end
    
    addonData.callback(1, function()
        local hasLoot, _ = CanLootUnit(destGuid)

        if hasLoot and IsInRaid() then
            sendAddonMessage("CAN_LOOT", destName, destGuid)
            return
        end
    end)
end

-- function events:READY_CHECK()
    -- if not addonData.isRaidLeader() then
        -- return
    -- end

    -- local raidMembers = addonData.getRaidMembers()
    -- local soulstoneIds = { 20762, 20707, 20764, 20765, 20763 }
    
    -- local anyWarlocks = false
    -- for i,raidMember in ipairs(raidMembers) do
        -- anyWarlocks = anyWarlocks or raidMember.class == "WARLOCK"
        -- if raidMember.class == "PRIEST" or raidMember.class == "PALADIN" then
            -- local j = 1
            -- local name,_,_,_,_,_,_,_,_,spellId = UnitBuff(raidMember.name, j)
            -- while name do
                -- if addonData.hasValue(soulstoneIds, spellId) then
                    -- return
                -- end
                -- j = j + 1
                -- name,_,_,_,_,_,_,_,_,spellId = UnitBuff(raidMember.name, j)
            -- end
        -- end
    -- end
    
    -- if anyWarlocks then
        -- addonData.sendRaidWarning("ACHTUNG: Kein Seelenstein gesetzt")    
    -- end
-- end

-------------
-- API
-------------

local function addSimilarItemId(itemId, similarItemId)
    local similarIds = similarItemIds[itemId]
    if not similarIds then
        similarItemIds[itemId] = {}
        similarIds = similarItemIds[itemId]
    end
    
    similarIds[#similarIds + 1] = similarItemId
end

addonData.addSimilarItemId = addSimilarItemId

local function tableFindPlayer(tab, name)
    for index, value in ipairs(tab) do
        if value.player == name then
            return value
        end
    end

    return nil
end

local function showPlayerInstallBase()
    if not IsInRaid() then
        return
    end

    local playerName, _ = UnitName("player")

    sendAddonMessage("DISCOVER")

    addonData.callback(1, function()
        local cols = {
            { ["name"] = "Name", ["width"] = 100 },
            { ["name"] = "Version", ["width"] = 75 },
        }
    
        local players = addonData.getRaidMembers()
        local data = {}

        for i, val in pairs(players) do
            local color = addonData.colorizePlayer(val.name)
            local playerCell = { value = val.name, color = color }
            local versionCell = { value = "MISSING" }

            local playerData = tableFindPlayer(playersWithAddon, val.name)
            if playerData then
                playerCell.value = playerData.player
                versionCell.value = playerData.version
            end

            table.insert(data, { cols = { playerCell, versionCell } })
        end

        addonData.createPlayerFrame(cols, data)
        playersWithAddon = {}
    end)
end

addonData.showPlayerInstallBase = showPlayerInstallBase

local function showItemList(itemId, location)
    if not IsInRaid() then
        return
    end

    sendAddonMessage("DISCOVER_ITEM", location or "inventory", itemId)

    addonData.callback(1, function()
        local cols = {
            { ["name"] = "Player", ["width"] = 120, ["align"] = "CENTER" },
            { ["name"] = "Quantity", ["width"] = 80, ["align"] = "CENTER" }
        }
    
        local players = addonData.getRaidMembers()
        local data = {}

        for i, val in pairs(players) do
            local playerData = tableFindPlayer(itemTrackingTable, val.name)
            local color = addonData.colorizePlayer(val.name)

            if not playerData then
                playerData = { player = val.name, quantity = "?" }
            end

            table.insert(data, {
                cols = {
                    { value = val.name, color = color },
                    { value = playerData.quantity }
                }
            })
        end

        local item = Item:CreateFromItemID(itemId);
        item:ContinueOnItemLoad(function ()
            local itemName, itemLink = GetItemInfo(itemId)
            addonData.createItemFrame(itemName, cols, data)
            itemTrackingTable = {}
        end)
    end)
end

addonData.showItemList = showItemList

local function setMaintank(name1, name2, name3)
    sendAddonMessage("SET_MT", name1, name2, name3)
end

addonData.setMaintank = setMaintank

------------------------
-- Register to events
------------------------

local frame = CreateFrame("FRAME", "MONO_EventFrame")

frame:SetScript("OnEvent", function(self, event, ...)
	events[event](self, ...)
end)

for k, v in pairs(events) do
    frame:RegisterEvent(k)
end